"""
Compiling utils for the testbench

Author: Daniel Perdices <daniel.perdices at uam.es>
"""
import subprocess
import tempfile

class CompileError(ValueError):
    pass

def compile_program(asm_code:str):
    """Compiles a program generating data and instructions memories.

    Args:
        asm_code (str): The ASM code

    Returns:
        (str, str): (data, instructions) memories
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        asm_filename = f"{tmpdirname}/program.s"
        obj_filename = f"{tmpdirname}/program.o"
        with open(asm_filename, "w") as f:
            f.write(asm_code)


        asm2obj(asm_filename, obj_filename)
        data, instruct =  obj2data(obj_filename), obj2text(obj_filename)
        return data, instruct


def asm2obj(filename:str, objname:str):
    """Generates the object file from the assembly file.

    Args:
        filename (str): path to the assembly file
        objname (str): path to the object file

    Raises:
        CompileError: Compiling error
    """
    command = ["mips-linux-gnu-as"] 
    command += ["-g2", "-g", "--gdwarf2", "-mips32", "-O0", "-o", objname, filename]
    proc = subprocess.run(command, capture_output=True)
    if proc.returncode != 0:
        raise CompileError(str(proc.stderr, encoding="utf8"))

def obj2data(objname:str, dataname=None):
    """Obtains the data memory from the object file.

    Args:
        objname (str): Object file path.
        dataname (str, optional): Data file path. Defaults to None.

    Raises:
        CompileError: Compiling error.

    Returns (only if dataname is None):
        str: data memory
    """
    command = ["mips-linux-gnu-objdump"]
    command += ["--full-contents",  objname]
    command += ["-j", ".data"]
    proc = subprocess.run(command, capture_output=True)
    
    if proc.returncode != 0:
        raise CompileError(str(proc.stderr, encoding="utf8"))
    
    lines = parse_full_contents(str(proc.stdout, encoding="utf8"))
    if dataname:
        with open(dataname, "w") as f:
            f.write("\n".join(lines))
    else:
        return "\n".join(lines)

def obj2text(objname, dataname=None):
    """Obtains the instruction memory from the object file.

    Args:
        objname (str): Object file path.
        dataname (str, optional): Instruction file path. Defaults to None.

    Raises:
        CompileError: Compiling error.

    Returns (only if dataname is None):
        str: instruction memory
    """
    command = ["mips-linux-gnu-objdump"]
    command += ["--full-contents",  objname]
    command += ["-j", ".text"]
    proc = subprocess.run(command, capture_output=True)
    
    if proc.returncode != 0:
        raise CompileError(str(proc.stderr, encoding="utf8"))
    lines = parse_full_contents(str(proc.stdout, encoding="utf8"))
    if dataname:
        with open(dataname, "w") as f:
            f.write("\n".join(lines))
    else:
        return "\n".join(lines)

def obj2commented(objname, commentedversion=None):
    """Obtains the commented instruction memory from the object file.

    Args:
        objname (str): Object file path.
        commentedversion (str, optional): Instruction file path. Defaults to None.

    Raises:
        CompileError: Compiling error.

    Returns (only if dataname is None):
        str: commented instruction memory
    """
    command = ["mips-linux-gnu-objdump"]
    command += ["-S",  objname]
    command += ["-j", ".text", "--source-comment=##  "]
    proc = subprocess.run(command, capture_output=True)
    
    if proc.returncode != 0:
        raise CompileError(str(proc.stderr, encoding="utf8"))
    if commentedversion:
        with open(commentedversion, "wb") as f:
            f.write(proc.stdout)
    else:
        return str(proc.stdout, encoding="utf8")

def parse_full_contents(text):
    """Parses the output of the objdump command.

    Args:
        text (str): Output of the objdump command.

    Returns:
        str: Formatted output.
    """
    res = []
    parsing = False
    for line in text.split("\n"):
        if ("Contents of section" in line) or ("Contenido de la" in line):
            parsing = True
            continue
        if parsing and len(line) > 0:
            line = line.strip()
            pieces = line.split(" ")
            address = int(pieces[0], base=16)
            for i in range(4):
                res.append("%08x\t%s" % ((address+i*4, pieces[i+1])))
    return res



from setuptools import setup, find_packages

VERSION = "0.5.4" 
DESCRIPTION = "A MIPS32 testbench in Python based on cocotb"
LONG_DESCRIPTION = "A MIPS32 testbench in Python based on cocotb"

setup(
        name="cocotb_mips32", 
        version=VERSION,
        author="Daniel Perdices",
        author_email="<daniel.perdices@uam.es>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        entry_points = {
            "console_scripts": ["cocotb-mips32 = cocotb_mips32.cli:main"],
        },
        install_requires=["cocotb", "docker"], # add any additional packages that 
        
)


import subprocess
import os
import tempfile
import argparse
import docker

def main():
    with tempfile.TemporaryDirectory(prefix="cocotb-mips32") as tmp_dir:
        parser = argparse.ArgumentParser(description="MIPS32 test suite")
        parser.add_argument("-m","--module", help="Python module", required=True)
        parser.add_argument("-t","--toplevel", help="Top Level entity name", default="processor")
        parser.add_argument("-s","--simulator", help="Simulator (questa or docker-questa)", default="questa")
        parser.add_argument("-d", "--rtl-dir", help="Path to the RTL folder with the vhdl source files", required=True)
        parser.add_argument("--stdout", help="stdout/stderr of the simulator", default=None)
        parser.add_argument("--docker-image", help="Docker image with cocotb and questa", default="cocotb-questa")

        args = vars(parser.parse_args())
        module, toplevel, simulator, rtl_dir = args["module"], args["toplevel"], args["simulator"], args["rtl_dir"]
        file_stdout, docker_image = args["stdout"], args["docker_image"]

        if simulator == "docker-questa":
            print("Running in docker")
            real_rtl_dir = os.path.abspath(rtl_dir)
            rtl_dir = "/rtl_dir"
            real_tmp_dir = os.path.abspath(tmp_dir)
            tmp_dir = "/tmp_dir"
            client = docker.from_env()
            fli_lib = client.containers.run(docker_image, command="cocotb-config --lib-name-path fli questa", auto_remove=True).decode("utf8").replace("\n", "")
        else:
            fli_lib = subprocess.run(["cocotb-config", "--lib-name-path", "fli", "questa"], stdout=subprocess.PIPE).stdout.decode("utf8").replace("\n", "")
        do_script = f"""# Autogenerated file
        onerror {{
        quit -f -code 1
        }}
        if [file exists {tmp_dir}/work] {{vdel -lib {tmp_dir}/work -all}}
        vlib {tmp_dir}/work
        vmap -c
        vmap work {tmp_dir}/work
        vcom -work work +acc {rtl_dir}/*.vhd
        vsim  -onfinish exit -foreign "cocotb_init {fli_lib}" work.{toplevel}
        run -all
        quit"""


        VSIM_PATH="vsim"
        env=os.environ.copy()
        env["MODULE"] = module
        env["TOPLEVEL"] = toplevel
        env["TOPLEVEL_LANG"] = "vhdl"
        cmd = [VSIM_PATH]
        cmd += ["-c", "-64"]

        if file_stdout:
            stdout, stderr = subprocess.PIPE, subprocess.STDOUT
        else:
            stdout, stderr = None, None
        
        if simulator == "docker-questa":
            vols = {
                real_tmp_dir: {
                    "bind": tmp_dir, "mode": "rw"
                },
                real_rtl_dir: {
                    "bind": rtl_dir, "mode": "ro"
                },
                os.getcwd(): {
                    "bind": "/cwd", "mode": "rw"
                }
            }
            print(env["LM_LICENSE_FILE"])
            env["PATH"] = f"{env['PATH']}:/opt/questasim/linux_x86_64"

            container = client.containers.run(docker_image, command="/bin/bash", stdout=True, volumes=vols, stderr=True, 
                            auto_remove=False, environment=env, detach=True, stdin_open=True, 
                            cap_add=["SYS_PTRACE"], security_opt=["seccomp=unconfined"])
            container.exec_run(["pip3", "install", "-U", "cocotb-mips32"])
            container.exec_run(cmd)
            # Writing to stdin
            input_socket = container.attach_socket(
                params={
                    "stdin":True,
                    "stdout":True,
                    "stderr":True,
                    "stream":True,
                },
            )
            input_socket._sock.sendall(do_script.encode("utf8"))
            input_socket._sock.close()
            container.stop()
            container.wait()
            print(container.logs().decode("utf8"))
            container.remove()
            
        else:
            proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, env=env, stderr=stderr, stdout=stdout)
            stdout, stderr = proc.communicate(do_script.encode("utf8"))
            if file_stdout:
                with open(file_stdout, "ab") as f:
                    f.write(stdout)
                    
if __name__ == "__main__":
    main()



from typing import Text
from cocotb.triggers import Timer, Edge, First, RisingEdge
import cocotb
import random
from cocotb_mips32.utils.compiling import compile_program

def to_int(a, default=0):
    """Converts a value to an integer."""
    try:
        return int(a)
    except:
        return default

    
class MIPS32ProcessorTest(object):
    CLOCK_CYCLE = 10 # ns
    data = {}
    instructions = {}
    clock_ticking = True
    data_text = None
    instructions_text = None
    asm_code = None
    asm_file = None
    initialized = False
    log_file = None
    result = "tbd"
    error = None
    t = 0
    def __init__(self) -> None:
        # If asm_file is present, read the asm code from the file
        if self.asm_file:
            with open(self.asm_file, "r") as f:
                self.asm_code = f.read()

        # If asm_code is present, compile it and generate text and data segments
        if self.asm_code:
            data, instruct = compile_program(self.asm_code)

            self.data_text = self.data_text if self.data_text else data
            self.instructions_text = self.instructions_text or instruct

        # If data segment is present, load it
        if self.data_text:
            self.data = self.data if self.data else self.memload_from_str(self.data_text)
        
        # If instructions segment is present, load it
        if self.instructions_text:
            self.instructions = self.instructions if self.instructions else self.memload_from_str(self.instructions_text)
        super().__init__()

    async def clock(self, uut):
        """Clock co-routine."""
        self.CLOCK_CYCLE = 2*(self.CLOCK_CYCLE // 2)
        self.log.debug(f"Starting clock (clock cycle={self.CLOCK_CYCLE}ns)")
        while self.clock_ticking:
            uut.Clk.value = 0
            await Timer(self.CLOCK_CYCLE // 2, units="ns")
            uut.Clk.value = 1
            await Timer(self.CLOCK_CYCLE // 2, units="ns")
            self.t += 1

    async def read_inst_mem(self, uut):
        """Async read instruction memory co-routine."""
        while self.clock_ticking:
            address = to_int(uut.IAddr.value)
            if address not in self.instructions:
                uut.IDataIn.value = 0
            else:
                uut.IDataIn.value = self.instructions[address]
            await First(Edge(uut.IAddr), Timer(self.CLOCK_CYCLE, units="ns"))

    async def read_data_mem(self, uut):
        """Async read data memory co-routine."""
        while self.clock_ticking:
            address = to_int(uut.DAddr.value)
            if address not in self.data:
                uut.DDataIn.value = 0
            else:
                uut.DDataIn.value = self.data[address]
            await First(Edge(uut.DAddr), Timer(self.CLOCK_CYCLE, units="ns"))

    async def write_data_mem(self, uut):
        """Sync write data memory co-routine."""
        while self.clock_ticking:
            await RisingEdge(uut.Clk)
            address = to_int(uut.DAddr.value)
            if uut.DWrEn.value == 1:
                self.data[address] = int(uut.DDataOut.value)
            
    async def wait_cycles(self, n):
        """Waits for n clock cycles."""
        await Timer(n*self.CLOCK_CYCLE, units="ns")
        return None


    @property
    def PC(self):
        return self.uut.IAddr

    @property
    def regs(self):
        return self.Regs(self.uut)

    def Regs(self, uut):
        return uut.RegsMIPS.regs

    def randomized_regs(self, uut, lower_limit=1, upper_limit=2**31-1):
        """Sets a randomized register set."""
        for i in range(1, 32):
            self.Regs()[i].value = random.randint(lower_limit, upper_limit)

    def assertRegEqual(self, reg, val):
        assert self.regs[reg] == val, "Register {} ({}) is not equal to {}".format(reg, self.regs[reg], val)

    def assertPCEqual(self, val):
        assert self.PC == val, "PC ({}) is not equal to {}".format(self.PC, val)

    def get_regs_as_dict(self, uut):
        """Returns the registers as a dictionary."""
        res = {}
        regs = self.Regs(uut)
        for i in range(32):
            res[i] = to_int(regs[i].value)
        return res

    def compare_regs(self, regs1, regs2):
        """Compares two register dicts."""
        for i in range(32):
            if regs1[i] != regs2[i]:
                return False
        return True

    def reset_processor(self, uut):
        """Resets the processor."""
        uut.Reset.value = 1
        uut.IDataIn.value = 0
        uut.DDataIn.value = 0
        uut.Clk.value = 0

    async def init_processor(self, uut):
        """Initializes the processor."""
        if self.initialized:
            return
        self.uut = uut
        self.log = self.uut._log
        self.reset_processor(uut)
        await self.wait_cycles(1)
        uut.Reset.value = 1
        cocotb.fork(self.clock(uut))
        cocotb.fork(self.read_data_mem(uut))
        cocotb.fork(self.read_inst_mem(uut))
        cocotb.fork(self.write_data_mem(uut))
        await self.wait_cycles(3)
        self.t = 0
        uut.Reset.value = 0
        self.initialized = True

    def memload_from_str(self, str):
        """Loads a memory (as a dict) from a string."""
        res = {}
        for line in str.split("\n"):
            if line[0] == "#":
                continue
            pieces = line.split("\t")
            if len(pieces) == 2:
                res[int(pieces[0], base=16)] = int(pieces[1], base=16)
        return res

    def log_result(self):
        """Logs the result to the file"""
        if self.log_file:
            with open(self.log_file, "a") as f:
                f.write(f"{self.__class__.__name__}\t{self.result}\n")

    def register_test(self):
        """Registers the test in the module."""
        async def wrapper(uut):
            await self.init_processor(uut)
            try:
                await self.main(uut)
                self.result = "correct"
            except AssertionError as e:
                self.result = "error"
                self.error = e
            except Exception as e:
                self.result = "exec_error"
                self.error = e
            
            self.log_result()
            if self.error:
                raise self.error

        f = wrapper
        f.__name__ = self.__class__.__name__
        f.__qualname__ = self.__class__.__name__
        f.__doc__ = self.__doc__ or self.asm_code
        globals()[self.__class__.__name__] = cocotb.test()(f)
        return globals()[self.__class__.__name__]

    async def main(self, uut):
        self.log.info(f"{to_int(uut.IAddr.value)}")
        await self.wait_cycles(3)
        self.log.info(f"{to_int(uut.IAddr.value)}")
        self.clock_ticking = False
        print("End of test")
